//created by Alex Hartley
//this script manages Item visibility for ikfk setups with the correct names or tags
//-----------------
// naming rules: "Owner_Type_Group_Name_[match][switchdefs]"
// Owner= "CharacterName" -no spaces or underscores
// valid Types= "IK", "FK", "Bones", "IKFKSwitch" 
// Group = "String" that all controllers that are part of the same IKFK 'system' share -no spaces or underscores
// Name = "String" that distinguishes this controller -no spaces or underscores multiple names are fine
// match = "_match" at the end of the name tells script that this item is for ikfk matching
// switchdefs = "@Position.Y(0,1)" this name part when applied to a switch tells the script what channel to use on the switch. 
// default switchdef is "@Position.Y(0,1)"


//-----------------

@version 1.0
@warnings
@script master
@name ikfkMaster

//-----------------
//Global Variables
cChannelShow,
t2,
color,
delayedUpdate,
updated,
newSelection,
storedSelect,
storedWarning,
storedFrame,
processingSwitches,
poseMatchChoice,
warningChoice,
typeChoice,
working,
userSelect,
switchType,
switchInvert,
switchChannels,
switchMin,
switchMax,
affectedOwners,
affectedSwitches,
affectedControllers,
currentControllers,
selectedControllers,
controllerLocalRotations,
controllerLocalPositions,
controllerWorldPositions,
controllerMatchItemIDs,
controllerGroups,
controllerTypes,
conciseOwners,
controllerOwners,
controllerNulls,
controllerBones,
controllerGenus,
controllerIDs,
controllerNames,
lastFrame,
frame,
fps,
live,
time, 
scene;

//-----------------

create
{
	t2 = time();
    color=<77,108,139>;
    typeChoice=recall("ikfkTypeChoice");
    if (!typeChoice)
        typeChoice=1;
    warningChoice=recall("ikfkWarningChoice");
    if (!warningChoice)
        warningChoice=1;
    poseMatchChoice=recall("ikfkPoseMatchChoice");
    if (!poseMatchChoice)
        poseMatchChoice=0;
	scene = Scene();
	fps=scene.fps;
    if (scene.loadinprogress==1)
    	delayedUpdate=true;
    else
    	updateFromScene();	
    
}

// listens for changes from layout
process: event, command
{
    time = scene.currenttime;
 	if (delayedUpdate==true)
 	{
 		if(scene.loadinprogress==0)
 		{
 			updateFromScene();
 			delayedUpdate=false;
 		} 	
 	}
 
    //parse the command
    command = parse(" ", string(command));
    // if the user tries to move to another frame show a warning if it exists.
    if(command[1]== "GoToFrame" || command[1]== "NextFrame" || command[1]== "PreviousFrame"|| command[1]== "GoToFirstFrame"|| command[1]== "GoToLastFrame"| command[1]== "NextKey"| command[1]== "PreviousKey")
    {
        if(storedWarning && warningChoice==1 && poseMatchChoice!=1)
        {
			working=true;
			storedWarning=nil;
			storedWarning=checkMatch();
			if(storedWarning)
			{
				warn(storedWarning);
				GoToFrame(storedFrame);
				working=false;
			}
			else
				updateVisibility();				
        }
        else
        	updateVisibility();	
    }
    else if(command[1]== "PlayForward"|| command[1]== "PlayBackward")
    {
     if(storedWarning && warningChoice==1 && poseMatchChoice!=1)
        {
            warn(storedWarning);
			storedWarning=nil;
			storedWarning=checkMatch();
			if(storedWarning)
			{
				working=true;
				Pause();
				GoToFrame(storedFrame);
				working=false;
			}			
        }
    }
    else if(!working)
    {
    	drawWarnings();
    }
    
	    
    if( !reqisopen() )
        options();
    // track the user moving controllers. if the user moves an item set the flag and store a warning.
    if(typeChoice==1 || poseMatchChoice)
    {
        if(!working)
        {
            if(command[1]=="Position" || command[1]=="Rotation"|| command[1]=="Scale" || command[1]=="AddPosition" || command[1]=="AddRotation"|| command[1]=="AddScale")
            {	
                if(!storedWarning)
                {
                	mismatch=false;
                    editTime=time;
                    color=<255,255,0>;
              //      requpdate();
                    storedFrame = integer(time*fps);
                    storedWarning = "You have made changes. Don't forget to sync IK/FK.";
                }               
            }
        }   
    }   
}

// Interface Code
options
{
    if( reqisopen() )
        reqend();
    else
    {
        reqbegin("ikfkMaster");
        reqsize(510,36);
        

        cKeyType = ctlchoice("",typeChoice,@"Step","Linear","TCB"@);
        ctlposition(cKeyType,4,8,122,20);
        ctlrefresh(cKeyType,"refreshType");

        
        cPoseMatch = ctlpopup("", poseMatchChoice,@ "Match None", "Match Chain", "Match Pose" @);
        ctlposition(cPoseMatch,128,8,100,20);
        ctlrefresh(cPoseMatch,"refreshPoseMatch");  
 //       ctlvisible(cKeyType,"showPoseMatchCheckbox",cPoseMatch);

        cWarning = ctlbutton("warn",32,"toggleWarning");
        ctlposition(cWarning,236,8,32,20); 
//        ctlvisible(cKeyType,"showWarningPopup",cWarning);

        
        cSwitch = ctlbutton("IK/FK Switch", 100,"switchSelected");
        ctlposition(cSwitch,276,8,100,20);   

        cSync = ctlbutton("Sync IK/FK", 100,"matchPose");
        ctlposition(cSync,376,8,100,20);   
       	
		cChannel = ctlbutton("C",16, "openGraphEditor");
		ctlposition(cChannel,484,8, 18,20);

		cChannelShow = ctlcheckbox("", 0);
		ctlposition(cChannelShow,584,8, 0,0);
		ctlactive(cChannelShow,"showChannelButton",cChannel);    	         
        
        reqopen();
        
    }
}
showChannelButton: value
{
return(value);
}
drawWarnings
{
	// draw border around C button if not full fk or IK
	count=nil;
	hasSwitch=false;
	t=7;
	b=28;
	l=483;
	r=502;
	l1=275;
	r1=476;
	l2=235;
	r2=268;

	blue=<77,108,139>;
	orange=<255,128,0>;	
	yellow=<255,255,0>;	
	darkyellow=<128,128,0>;	
	green=<0,255,0>;	
	selection=scene.getSelect();
	for(i=1;i<=switchChannels.size();i++)
	{
		if(selection)
		{
			if(integer(selection[1].id) == integer(controllerIDs[i]) && switchChannels[i])
			{
				hasSwitch=true;
				if(switchType[i]=="Rotation")
				{
					max=switchMax[i]*3.14159/180;
					min=switchMin[i]*3.14159/180;       
				}
				else
				{
					max=switchMax[i];
					min=switchMin[i];       
				}
					count=i;
			}
		}
		
	}
	if (warningChoice==1)
	{
		drawline(darkyellow,l2,t,r2,t);
		drawline(darkyellow,l2,b,r2,b);
		drawline(darkyellow,l2,t,l2,b);
		drawline(darkyellow,r2,t,r2,b);	
	}
	else
	{
			drawerase(l2,t,1,b - t);
			drawerase(r2,t,1,b - t);
			drawerase(l2,t,r2 - l2,1);
			drawerase(l2,b,r2 - l2 + 1,1);
	}


	if(storedWarning && typeChoice==1 && warningChoice!=0)
		wcolor=yellow;
	else
		wcolor=blue;
	drawline(wcolor,l1,t,r1,t);
	drawline(wcolor,l1,b,r1,b);
	drawline(wcolor,l1,t,l1,b);
	drawline(wcolor,r1,t,r1,b);

	if(hasSwitch && switchChannels[count])
	{	
		if(reqisopen() && !getvalue(cChannelShow) && reqisopen())
		{
			setvalue(cChannelShow,1);
			requpdate();	
		}		
		if(switchChannels[count].keyExists(time))
		{
			drawline(green,l,t,r,t);
			drawline(green,l,b,r,b);
			drawline(green,l,t,l,b);
			drawline(green,r,t,r,b);		
		
		}
		else if(switchChannels[count].value(time) != max && switchChannels[count].value(time) != min)
		{
			drawline(orange,l,t,r,t);
			drawline(orange,l,b,r,b);
			drawline(orange,l,t,l,b);
			drawline(orange,r,t,r,b);		
		
		}
		else
		{
			drawerase(l,t,1,b - t);
			drawerase(r,t,1,b - t);
			drawerase(l,t,r - l,1);
			drawerase(l,b,r - l + 1,1);
		}	

	}
	else
	{
		if(reqisopen() && getvalue(cChannelShow) && reqisopen())
		{
			setvalue(cChannelShow,0);
			requpdate();		
		}
		drawerase(l,t,1,b - t);
		drawerase(r,t,1,b - t);
		drawerase(l,t,r - l,1);
		drawerase(l,b,r - l + 1,1);
	}
}
showPoseMatchCheckbox: value
{
    if (value != 1)
        return(true);
    else
        return(false);  
}
showWarningPopup: value
{
    if (value == 1)
        return(true);
    else
        return(false);  
}
refreshPoseMatch: val
{
    poseMatchChoice=val;
    store("ikfkPoseMatchChoice", val);
}
toggleWarning
{
	if(warningChoice==0)
		warningChoice=1;
	else
	{
		warningChoice=0;	
		storedWarning=nil;
	}

    store("ikfkWarningChoice", val);
}   
refreshType: val
{
    typeChoice = val;
    store("ikfkTypeChoice", val);
}
warningState
{
info("working =" + string(working) + ", storedWarning = "+ string(storedWarning)); 
}
openGraphEditor
{
	selection=nil;
	selection = scene.getSelect();
	userSelect = selection;
	inArray = nil;
    inArray= findinArray(selection);
    for(i=1;i<=inArray.size();i++)
 	CommandInput("GE_ClearBin");
	CommandInput("GE_OpenWindow 1");
	first=true;
	count=1;
    for(i=1; i<=inArray.size(); i++)
    {
        if(inArray[i])
        {
			name=nil;
        	for(j=1;j<=selection.size();j++)
        	{
        		object=nil;
        		
        		if(integer(controllerIDs[i])==integer(selection[j].id))
        		{
					if(selection[j].isBone())
					{
						EditObjects();
						object = Mesh(0);
						name= object.name +"."+ selection[j].name;
					}
					else
						name= selection[j].name;

        		}
        	}
        	if(name)
        	{
				 command[count] = "GE_SetEnv" +" \""+ switchChannels[i].parent.name +"."+ switchChannels[i].name +"\" "+ "1";
				 count++;
        	}						
			for(j=1;j<=controllerGroups.size();j++)
			{
				if(controllerGroups[j])
				{
					if(controllerGroups[j] == controllerGroups[i])
					{
						if(controllerTypes[j]=="IK" || controllerTypes[j]=="FK" || controllerTypes[j]=="Bones")
						{
							item=nil;
							SelectItem(integer(controllerIDs[j]));
							item=scene.getSelect();
							if(!item[1].locked)
							{
							CommandInput("GE_GetLayoutSel 1");
							}
						}
					}				
				}	
			}
        }
    }
    for(i=1;i<=command.size();i++)
    {
		CommandInput(command[i]);
    }    
	restoreUserSelect();
}
channelTest
{
    selection = scene.getSelect();
    channel[1] = selection[1].firstChannel();
    channel[2]  = selection[1].nextChannel();
    channel[3]  = selection[1].nextChannel();
    channel[4]  = selection[1].nextChannel();
    channel[5]  = selection[1].nextChannel();
    channel[6]  = selection[1].nextChannel();
    

    for(i=1; i<= channel.size(); i++)
    {
     if (i > 3)
        value[i]= deg(channel[i].value(time));
     else
        value[i]= channel[i].value(time);
        
        info(string(channel[i].name) + " " + string(value[i]));
    }
    
    
}
listChannels
{
for (i=1; i<=switchChannels.size(); i++)
    if(switchChannels[i])
        info(switchChannels[i]);
}
listItems
{
    for(i=1; i<=controllerNames.size(); i++)
    {
        info(controllerNames[i] + " - ID: " + controllerIDs[i] + " - group: " + controllerGroups[i]);
    }
}
showInfo
{
    selection = scene.getSelect();
    // find the selected controls in the array and show info
        
    for(i=1; i<=controllerNames.size(); i++)
    {
        for(j=1; j<=selection.size(); j++)
        {
            if(string(controllerIDs[i]) == string(selection[j].id) )
                info(controllerNames[i] + " - ID: " + controllerIDs[i] + " - group: " + controllerGroups[i] + " - switch state: " + string(switchChannels[i].value(time)) + " - position: " + string(controllerLocalPositions[i]) + " - rotation: " + string(controllerLocalRotations[i])  ); 
        }
    }
}
findAffectedSwitches
{
    selectedControllers = nil;
    affectedSwitches = nil;
    selection = userSelect;
    // find the selected controls in the array
    
    selectedControllers= findinArray(selection);
        
    //find all affected controllers and Switches in the array
    for(i=1; i<=controllerNames.size(); i++)
    {
        for(j=1; j<=selectedControllers.size(); j++)
        if(selectedControllers[j])
        {
            {
                if(controllerOwners[i] == controllerOwners[j] && controllerGroups[i] == controllerGroups[j] && controllerTypes[i] == "IKFKSwitch" )
                    affectedSwitches[i] = integer(controllerIDs[i]);                                            
            }        
        }        
    }
    count=0;
    for(i=1; i<=affectedSwitches.size(); i++)
    {
        if(affectedSwitches[i])
            count++;
    }
    if(count==0)    
        error("there are no Switches associated with your selection");  
    return(affectedSwitches);
}
findinArray: selection
{
    working= true;
    inArray=nil;
    for(i=1; i<=controllerNames.size(); i++)
    {
        for(j=1; j<=selection.size(); j++)
        {
        	if(controllerIDs[i])
        	{
				if(string(controllerIDs[i]) == string(selection[j].id) )
				{
					inArray[i] = integer(controllerIDs[i]); 
					//info(controllerNames[i]);            
				}
        	}
        }
    }
    return(inArray);
    working= false;
}
switchSelected
{
    working=true;
	userSelect=nil;
	userSelect=scene.getSelect();
	updated=nil;
	selectedControllers=nil;
	updated = updateFromScene(); 
    // get the affected items
    affectedSwitches = findAffectedSwitches();
    if(poseMatchChoice !=1)
        matched = matchPose();
    working=true;
    if(!selectedControllers)
    	selectedControllers=findinArray(userSelect);  
    //toggle the switches
    for (i=1; i <=affectedSwitches.size(); i++)
    {
        if(affectedSwitches[i] != nil)
        {
            if(switchType[i]=="Rotation")
            {
                max=rad(switchMax[i]);
                min=rad(switchMin[i]);      
            }
            else
            {
                max=switchMax[i];
                min=switchMin[i];       
            }
            if(!switchInvert[i])
            {
                if(switchChannels[i].value(time)<=min)
                    key = switchChannels[i].createKey(time,switchMax[i]);
                else        
                    key = switchChannels[i].createKey(time,switchMin[i]);           
            }
            else
            {
                if(switchChannels[i].value(time)>=min)
                    key = switchChannels[i].createKey(time,switchMax[i]);
                else        
                    key = switchChannels[i].createKey(time,switchMin[i]);                       
            }

            if(typeChoice == 1)
                switchChannels[i].setKeyCurve(key, CHAN_STEPPED);
            if(typeChoice == 2)
                switchChannels[i].setKeyCurve(key, CHAN_LINEAR);
            if(typeChoice == 3)
            {
                switchChannels[i].setKeyCurve(key, CHAN_TCB);
                switchChannels[i].setKeyTension(key, 1);
            }
        }       
    }
    
    positions = getControllerPositions();
        
    // select an appropriate controller for user
	myItem=nil;
    for (i=1; i <=selectedControllers.size(); i++)
    {
        if(selectedControllers[i])
        {
            SelectItem(integer(controllerIDs[i]));
            item1=scene.getSelect();
            for(j=1; j<=controllerGroups.size(); j++)
            {
                if (controllerOwners[i]==controllerOwners[j] && controllerGroups[i]==controllerGroups[j])
				if(controllerTypes[i] != controllerTypes[j] && (controllerTypes[j]=="Bones" || controllerTypes[j]=="IK" || controllerTypes[j]=="FK" ))
                {
                    SelectItem(integer(controllerIDs[j]));
                    item2 = scene.getSelect();
                    if(!item2[1].locked)
                    {
                        myItemID=integer(item2[1].id);
                        if( item1[1].getWorldPosition(time) == item2[1].getWorldPosition(time) )
                           i = selectedControllers.size();                          
                    }
                    
                }
            }        
        }
	}
    updateVisibility(); 
    SelectItem(myItemID);
    updated=nil;
    working=false;  
    
}
matchPose
{
	//Make sure that pose matching is on.
	if(poseMatchChoice==1 )
		return(true);
				
    working=true;
    if(!updated)
    {
		userSelect = nil; 
		userSelect = scene.getSelect();    
        updated = updateFromScene();
    }
	//clear arrays
	affectedControllers = nil;
    selectedControllers = nil;
    affectedOwners = nil;
	affectedGroups = nil;
	BonesIDs=nil;
	NullsIDs=nil;

    
    //build arrays of affected groups and owners.
    selectedControllers = findinArray(userSelect);
    j=1;
    for(i=1;i<=selectedControllers.size();i++)
    {
    	if(selectedControllers[i])
    	{
			for(j=1;j<=controllerOwners.size();j++)
			{
				if(controllerOwners[i]==controllerOwners[j])
				{
					affectedOwners[j]=controllerOwners[j];
					if(controllerGroups.size()>=j && controllerGroups.size()>=i)
					{
						if(controllerGroups[i]==controllerGroups[j])
						{
							affectedGroups[j]=controllerGroups[j];

						}    		
					}
					
				}    		
			}
    	}		
    }    
    if(poseMatchChoice == 3)
    	affectedControllers = affectedOwners;
    if(poseMatchChoice == 2)
    	affectedControllers = affectedGroups;


	//clear arrays
    affectedOwners = nil;
	affectedGroups = nil;

    // separate the affected bones and nulls
    for(i=1; i<=affectedControllers.size(); i++)
    {
        if(affectedControllers[i] && controllerIDs[i] && controllerTypes[i] == "Bones" && controllerGenus[i] == 4 )    
            BonesIDs[i] = controllerIDs[i];
        if(affectedControllers[i] && controllerIDs[i] && controllerTypes[i] != "Bones")  
            NullIDs[i] = controllerIDs[i];              
    }
    
    //key all the nulls
    first= true;    
    for(i=1; i<=NullIDs.size(); i++)
    {
        key=nil;
        if(NullIDs[i] && first)
        {               
            SelectItem(integer(NullIDs[i]));
            first = nil;            
        }
        else if(NullIDs[i])
        {
            AddToSelection(integer(NullIDs[i]));
        }
    }
    CreateKey(time);
    
    // key all the  bones
    first= true;
    for(i=1; i<=BonesIDs.size(); i++)
    {
        if(BonesIDs[i] && first)
        {
            SelectItem(integer(BonesIDs[i]));
            first=nil;
        }
        else if (BonesIDs[i])
        {
            AddToSelection(integer(BonesIDs[i]));
        }   
    }
    CreateKey(time);
    //get the positions of the match items and assign them to their counterparts.
    positions = getControllerPositions();
    for (i=1; i<=affectedControllers.size(); i++)
    {
        if(affectedControllers.size()>=i && controllerMatchItemIDs.size()>=i)
        {
            if (controllerMatchItemIDs[i] && affectedControllers[i])
            {
//            	info(controllerNames[i]);
                for(j=1; j<=controllerMatchItemIDs.size(); j++)
                {
                    if (controllerMatchItemIDs[i] == controllerIDs[j])
                    {
                        controllerLocalPositions[i] = ( controllerLocalPositions[j] );
                        controllerLocalRotations[i] = ( controllerLocalRotations[j] );                     
                    }
                }
            }
        }      
    }
	
    // update the match items and key again.
    for(i=1; i<=affectedControllers.size(); i++)
    {
    	if(i<=controllerMatchItemIDs.size())
    	{
			if(controllerMatchItemIDs[i] && affectedControllers[i])
			{
				SelectItem(integer(controllerIDs[i]));
				Position(controllerLocalPositions[i]);
				Rotation(controllerLocalRotations[i]);
				CreateKey(time);
			}       
    	}
		
    }
     //clear arrays
	affectedControllers = nil;


    //restore the user selection
    for(i=1; i<=userSelect.size(); i++)
    {
        if (i == 1)
            SelectItem(userSelect[i].id);   
        else        
            AddToSelection(userSelect[i].id);
    }    
    //info(keyedGroups);
    BonesIDs=nil;
	NullsIDs=nil;
	if (storedWarning)
	{
		storedWarning=checkMatch();
		storedFrame = integer(time*fps);
	}    
    color=<77,108,139>;
    requpdate();
    working=false;
    updated=nil;
    return(true);	
}
updateFromScene
{
    userSelect = nil;
    userSelect = scene.getSelect();
    time = scene.currenttime;
    fps= scene.fps;
    frame = time*fps;
    //get arrays
    controllerNames = getcontrollerNames(); 
    controllerOwners = getControllerOwners();
    conciseOwners = getConciseOwners();
    controllerTypes = getControllerTypes();
    controllerGroups = getControllerGroups();
    controllerIDs = getControllerIDs();
    controllerGenus = getControllerGenus();

    // get mesh object agents and bone object agents
    getControllerNullsBones();
    
    // get switchChannels, switchMax, and switchMin Arrays
    getSwitchDefs();
    while(processing)
        processing=processingSwitches;
        
    
    //get positions and rotations
    positions = getControllerPositions();
    //store selection

    // shorten name
    for(i=1; i<=controllerNames.size(); i++)
    {
        controllerNames[i]= strright(controllerNames[i], controllerNames[i].size() - controllerIDs[i].size() - 1 );
    }
        controllerMatchItemIDs = getcontrollerMatchItemIDs();
	restoreUserSelect();
    updateVisibility();
    setup = false;
    return(true);
}
getcontrollerNames  
{
    scene = Scene();
    //get every object in the scene
    EditObjects();
    SelectAllObjects();
    selItems = scene.getSelect();
	controllerNames=nil;   
    //make list of object names
    count = 1;
    for(i=1; i<=selItems.size(); i++)
    {
    	name=nil;
        //check for valid tags first
        tag = nil;
        tag = getIKFKTag(selItems[i]);
        if (tag != nil)
            name = tag; 
        else //if there is no valid tag use the name     
        	name= selItems[i].name;
		nameParts = parse("_", name);
		if(nameParts.size() > 1)
		{
//			if(nameParts[2] == "FK" || nameParts[2] == "IK" || nameParts[2] == "Bones" || nameParts[2] == "IKFKSwitch" )
			if(!selItems[i].locked || strright(name,6)=="_match" || (nameParts[2] == "Bones" && !selItems[i].isBone()))
			{
				objectNames[count] = string(selItems[i].id)  + "_" + name ;  
				count++;
			}   
		}
    }
    
    
    //find the bone objects
    count2 = 1;
    for(i=1; i<=objectNames.size(); i++)
    {    
        if(strright(objectNames[i], 5) == "Bones")
        {
            nameParts = parse("_",objectNames[i]);
            id = integer(nameParts[1]);
            nameParts=nil;
            SelectItem(id);
            id = nil;
            mesh = Mesh(0);
            boneObjectNames[count2] = mesh.name;
            count2++;
        }
    }
        // lose the original object array
    selItems = nil;

    // tell me if the bone objects are missing and quit
    if (boneObjectNames == nil)
    {
        error("can't find properly named or tagged bone objects");
        return;
    }
    //add bones to the list of object names with Object name as first two nameParts    
    for(i=1; i<= boneObjectNames.size(); i++)   
    {
        EditObjects();
        SelectItem(boneObjectNames[i]);
        SelectAllBones();
        EditBones();
        selBones = scene.getSelect();
        for(j=1; j<=selBones.size(); j++)
        {  
            if(!selBones[j].locked)
            {
                //check for tags first
                tag = nil;
                tag = getIKFKTag(selBones[j]);
                if (tag != nil)
                {
                    objectNames[count] = string(selBones[j].id) + "_" + tag;
                    tag=nil;
                }
                else
                {
                    boneNames[j] = selBones[j].name;
                    objectNames[count] = string(selBones[j].id) + "_" + boneObjectNames[i] + "_" + boneNames[j];                
                }
                count++;                
            }

        }       
        selBones = nil ;
        boneNames = nil ;
    }
    // lose the original bone array


     
    
    // find switches in scene
    count = 1;
    for(i=1; i<=objectNames.size(); i++)
    {
        nameParts = parse( "_" , objectNames[i]);
        s = size(nameParts);
        if (s >= 3)
        {
            if (nameParts[3] == "IKFKSwitch")
            {
                switchNames[count] = objectNames[i];
                count++;
            }
        }
    }


    // tell me if the switches are missing
    if (switchNames == nil)
    {
        error("can't find switches");
        return;
    }
    // find the controller items by comparing each item name to switch names
    count = 1;
    for(i=1; i<=objectNames.size(); i++)
    {
        nameParts = parse("_",objectNames[i]);
        s = size(nameParts);
        if (s >= 3)
        {
            for(i2=1; i2<=switchNames.size(); i2++)
            {
                switchNameParts = parse("_", switchNames[i2]);
                if(switchNameParts[2] == nameParts[2] && switchNameParts[4] == nameParts[4])
                {
                    controllerNames[count] = objectNames[i];
                    count++;
                }   
            }
        }
    }
    
    
    //tell me the names 
    //( "controllers found: " + controllerNames.size());
//    return(controllerNames);
	return(objectNames);
}
getControllerIDs
{
    if(controllerNames == nil)
    return;
    
    else
    {   
        for(i=1; i<=controllerNames.size(); i++)
        {
            nameParts = parse("_", controllerNames[i]);
            array[i] = nameParts[1];
        }
    return(array);  
    }
}
getControllerNullsBones
{
    controllerNulls = nil;
    controllerBones = nil;
    for(i=1; i<=controllerIDs.size(); i++)
    {
        SelectItem(integer(controllerIDs[i]));
        selection = scene.getSelect();
        if(selection[1].isMesh() && !selection[1].locked && controllerTypes[i] != "IKFKSwitch")
        {
            controllerNulls[i] = selection[1]; 
        }
        
        if(selection[1].isBone() && !selection[1].locked)
        {
            controllerBones[i] = selection[1]; 
        }
        
    }   
}
getControllerOwners
{
    if(controllerNames == nil)
    return;
    
    else
    {   
        for(i=1; i<=controllerNames.size(); i++)
        {
            nameParts = parse("_", controllerNames[i]);
            array[i] = nameParts[2];
        }
    return(array);  
    }
}
getConciseOwners
{
	array=nil;
	for(i=1;i<=controllerOwners.size();i++)
	{
		for(j=1;j<=controllerOwners.size();j++)
		{
			if(controllerOwners[i]==controllerOwners[j])
			{
				if(!array)
				{
					array[1]=controllerOwners[j];
				}
				else
				{
					found=false;
					for(k=1;k<=array.size();k++)
					{
						if(array[k]==controllerOwners[j])
							found=true;
					}
					if(!found)
					{
						k++;
						array[k]=controllerOwners[j];
					}					
				}
			}
		}
	}
	return(array);
}
getControllerTypes
{
    if(controllerNames == nil)
      return;
    
    else
    {   
        for(i=1; i<=controllerNames.size(); i++)
        {
            nameParts = parse("_", controllerNames[i]);
            array[i] = nameParts[3];
        }    
    return(array);  
    }
}
getControllerGenus
{
    if(controllerNames == nil)
      return;
    
    else
    {   
        for(i=1; i<=controllerNames.size(); i++)
        {
            SelectItem(integer(controllerIDs[i]));
            item=scene.getSelect();
            array[i] = item[1].genus;
            item=nil;
        }    
    return(array);  
    }
}
getcontrollerMatchItemIDs
{
    array= nil;
    if(controllerNames == nil)
        return;    
    for (i=1; i<=controllerNames.size(); i++)
    {
        nameParts = parse( "#", controllerNames[i]);
        if(nameParts[1].size() != controllerNames[i].size())
            nameParts[1]=strleft(nameParts[1], nameParts[1].size() - 1 );
        for(j=1; j<=controllerNames.size(); j++)
        {
            if ( nameParts[1]+"_match" == controllerNames[j])
            {
                array[i] = controllerIDs[j];
            }       
        }
        nameParts= nil;
    }
    return(array);
}
getControllerGroups
{
    if(controllerNames == nil)
    return;
    
    else
    {   
        for(i=1; i<=controllerNames.size(); i++)
        {
			nameParts = nil;
			nameParts = parse("_", controllerNames[i]);
			if (nameParts.size()>=4)
				array[i] = nameParts[4];	
        }
    return(array);  
    }
}
getControllerPositions
{
	//	clear arrays
	controllerLocalPositions=nil;
	controllerLocalRotations=nil;
	controllerWorldPositions=nil;

	// fill arrays
    for(i=1; i<=controllerIDs.size(); i++)
    {
        SelectItem(integer(controllerIDs[i]));
        selection = scene.getSelect();
        controllerWorldPositions[i] = selection[1].getWorldPosition(time);

        x = selection[1].firstChannel();
        y = selection[1].nextChannel();
        z = selection[1].nextChannel();
        h = selection[1].nextChannel();
        p = selection[1].nextChannel();
        b = selection[1].nextChannel();
        
        controllerLocalPositions[i] = vector(string(x.value(time))+" "+string(y.value(time))+" "+string(z.value(time)));
        controllerLocalRotations[i] = vector(string(deg(h.value(time)))+" "+string(deg(p.value(time)))+" "+string(deg(b.value(time))));                
    }
	return(true);
}
getSwitchDefs
{   
    processingSwitches= true;
    for(i=1; i<=controllerTypes.size(); i++)
    {
        if(controllerTypes[i] == "IKFKSwitch")
        {
            //check if the name or tag contains a SwitchDef string like "@Position.X(0,1)"
            nameParts=parse("@",controllerNames[i]);
            if(nameParts.size()>1)
            {
                defs = parse("(,)",nameParts[2]);
                
                //check for errors in defs.
                if(defs.size()!= 3)
                    error("illegal switch definition for " + mesh.name, ". Should be format: '@Position.X(0,1)'");                              
                if (defs[1]!="Position.X" && defs[1]!="Position.Y" && defs[1]!="Position.Z" && defs[1]!="Rotation.H" && defs[1]!="Rotation.P" && defs[1]!="Rotation.B")
                    error("Channel label in Switch Def not valid for " + mesh.name, ". Should be format: '@Position.X(0,1)'");
                if (string(number(defs[2]))!= defs[2])
                    error("Min Value in Switch Def not valid for " + mesh.name, ". Should be format: '@Position.X(0,1)'");
                if (string(number(defs[3]))!= defs[3])
                    error("Min Value in Switch Def not valid for " + mesh.name, ". Should be format: '@Position.X(0,1)'");
            }
            else
                //use default defs if there are no defs.
                defs = @"Position.Y","0","1"@;
            //get Switchchannels        
            SelectItem(integer(controllerIDs[i]));
            mesh = Mesh(0);
            channels[i] = mesh.firstChannel(); 
            while(channels[i])
            {
                last if channels[i].name == defs[1];
                channels[i] = mesh.nextChannel(); 
            }
            //get SwitchMinandMax
            min[i] = number(defs[2]);
            max[i] = number(defs[3]);

            //store what type of Switch         
            if (strleft(defs[1],8) == "Position")
                type[i] = "Position";
            else
                type[i] = "Rotation";

            //check if the min max values are inverted                  
            if(min[i] > max[i])
                invert[i] = true;
            else
                invert[i] = false;
            
        }
        else
            channels[i] = nil;      
    }
    // Assign the Switch Defs to the whole group
    for(i=1; i<=controllerTypes.size(); i++)
    {
        for(j=1; j<=channels.size(); j++)
        {
            if(channels[j] && controllerGroups[i] == controllerGroups[j] && controllerOwners[i] == controllerOwners[j])
            {
                switchChannels[i] = channels[j];                
                switchMin[i] = min[j];                  
                switchMax[i] = max[j];  
                switchInvert[i] = invert[j];
                switchType[i] = type[j];                
            }           
        }           
    }
    channels=nil;
    min=nil;
    max=nil;
    defs=nil;   
    processingSwitches = false;
}
updateVisibility
{
	working=true;
    userSelect = nil;
    userSelect = scene.getSelect();
    
    // Show or Hide Nulls
    for(i=1; i<=switchChannels.size(); i++)
    {
    
        if (switchChannels[i])
        {
			if(switchType[i]=="Rotation")
			{
				max=switchMax[i]*3.14159/180;
				min=switchMin[i]*3.14159/180;       
			}
			else
			{
				max=switchMax[i];
				min=switchMin[i];       
			}
            if( controllerTypes[i] == "FK" )
            {
                SelectItem(integer(controllerIDs[i]));
				if(!Mesh(0).locked)
				{
					if(!switchInvert[i])
					{
						if(switchChannels[i].value(time) > min)
							ItemVisibility(8);
						else
							ItemVisibility(1);                                                       
					}
					else
					{
						if(switchChannels[i].value(time) < min)
							ItemVisibility(8);
						else
							ItemVisibility(1);                                                      
					}
				}
            }
            if( controllerTypes[i] == "Bones" )
            {
                SelectItem(integer(controllerIDs[i]));
                if(!switchInvert[i])
                {
                    if(switchChannels[i].value(time) > min)
                        ItemVisibility(2);
                    else
                        ItemVisibility(1);                                                       
                }
                else
                {
                    if(switchChannels[i].value(time) < min)
                        ItemVisibility(2);
                    else
                        ItemVisibility(1);                                                      
                }

            }              
            if( controllerTypes[i] == "IK" )
            {
                SelectItem(integer(controllerIDs[i]));
                if(!Mesh(0).locked)
                {
					if(!switchInvert[i])
					{
						if(switchChannels[i].value(time) < max)
							ItemVisibility(8);
						else
							ItemVisibility(1);                                                      
					}
					else
					{
						if(switchChannels[i].value(time) > max)
							ItemVisibility(8);
						else
							ItemVisibility(1);                                                      
					}   
                }
                                                
            } 
                  
        }       
    }
/*    // Show or Hide BOnes
    for(i=1; i<=switchChannels.size(); i++)
    {
        if (switchChannels[i])
        {
            controllerBones[i].select();
                if(!switchInvert[i])
                {
                    if(switchChannels[i].value(time) > min)
                        ItemVisibility(2);
                    else
                        ItemVisibility(1);                                                      
                }
                else
                {
                    if(switchChannels[i].value(time) < min)
                        ItemVisibility(2);
                    else
                        ItemVisibility(1);                                                      
                }                                       
        }               
    }   */
    //restore the user selection
    restoreUserSelect();
    working=false;
}
restoreUserSelect
{
    for(i=1; i<=userSelect.size(); i++)
    {
        if (i == 1)
            SelectItem(userSelect[i].id);   
        else        
            AddToSelection(userSelect[i].id);
    }   
}
getIKFKTag: item
{
    i=1;
    while(!done)
    {
        tagParts = nil;
        tag = item.getTag(i);
        if (tag == nil)
        {
            return(nil);
            done=true;
        }
        
        else
        {
            tagParts = parse("=",tag);
            if (tagParts[1] == "IKFK")
                return(tagParts[2]);        
        }
        i++;    
    }
}
checkMatch
{
	userSelect=scene.getSelect();
	//get the positions of the match items and assign them to their counterparts.
	mismatch=nil;
	controllerLocalPositions=nil;
	controllerLocalRotations=nil;
    getControllerPositions(); 
    for (i=1; i<=controllerIDs.size(); i++)
    {
        if(controllerMatchItemIDs.size()>=i)
        {
            if (controllerMatchItemIDs[i])
            {
                for(j=1; j<=controllerMatchItemIDs.size(); j++)
                {
                    if (controllerMatchItemIDs[i] == controllerIDs[j])
                    {
                        matchPositions[i] = controllerLocalPositions[j];
                        matchRotations[i] = controllerLocalRotations[j];                      
                    }
                }
            }
        }      
    }
    //find the controllers that do not match their match counterparts
    matchTolerance=0.0001;
    for(i=1;i<=controllerNames.size();i++)
    { 
        if(controllerMatchItemIDs.size()>=i)
        {
            if(controllerMatchItemIDs[i])
            {   
                 p1 = controllerLocalPositions[i];
                 p2 = matchPositions[i];  
                 r1 = controllerLocalPositions[i];
                 r2 = matchRotations[i];
                if(vmag(p1 - p2) > matchTolerance && vmag(r1 - r2) > matchTolerance)
                {
//                  info(item[1].name+" to "+match[1].name+": " + string(vmag(p1 - p2)) +", "+string(vmag(r1[1] - r2[1])) +", "+string(vmag(r1[2] - r2[2])));
                    mismatch[i] = controllerIDs[i];    
                }
            }
        }
        item=nil;
        match=nil;   
    }
    //generate concise list of out of sync groups.
    warnGroups=nil;
    for(i=1;i<=mismatch.size();i++)
    {
    	if(mismatch[i])
    	{
			for(j=1;j<=controllerGroups.size();j++)
			{
				if (controllerGroups[i]==controllerGroups[j] && controllerOwners[i]==controllerOwners[j])
				{
					if(!warnGroups)
					{
						warnGroups[1]=controllerOwners[j]+" "+controllerGroups[j];
					}
					else
					{
						found=nil;
						k=1;
						while(warnGroups.size()>=k)
						{
							if(warnGroups[k]==controllerOwners[j]+" "+controllerGroups[j])
								found=true;
							k++;
						}
						if(!found)
							warnGroups[k]=controllerOwners[j]+" "+controllerGroups[j];
					}
				}
			}
    	}
    }
    //clear arrays
    
    restoreUserSelect();
    // generate warnings
    if(warnGroups.size()>1)
    {
		storedWarning =  string(warnGroups.size()) + " chains have un-synced changes. Don't forget to sync IK/FK.";    
		return(storedWarning);
    }

    if(warnGroups.size()==1)
    {
		storedWarning =  warnGroups[1] + " has un-synced changes. Don't forget to sync IK/FK.";    
		return(storedWarning);
    }

	else	
	{
		storedWarning = nil;
		return(storedWarning);
	}

		

}



